import random

import numpy as np
import torch
import torch.nn as nn
from gymnasium import spaces
from tqdm import trange


class ConstantRewardEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.observation_space = {
            "agent_0": spaces.Discrete(1),
            "agent_1": spaces.Discrete(1),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [{"agent_0": np.array([[0]]), "agent_1": np.array([[0]])}]
        self.sample_actions = [
            {"agent_0": np.array([[0.2, 0.8]]), "agent_1": np.array([[0.8, 0.2]])}
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None]

    def step(self, action):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        reward = {"agent_0": 1, "agent_1": 0}  # Constant reward of 1
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        info = {}
        return observation, info


class ConstantRewardImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))}
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2, 0.8]]), "agent_1": np.array([[0.8, 0.2]])}
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None]

    def step(self, action):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        reward = {"agent_0": 1, "agent_1": 0}  # Constant reward of 1
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        info = {}
        return observation, info


class ConstantRewardContActionsEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.observation_space = {
            "agent_0": spaces.Discrete(1),
            "agent_1": spaces.Discrete(1),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [{"agent_0": np.array([[0]]), "agent_1": np.array([[0]])}]
        self.sample_actions = [
            {"agent_0": np.array([[0.0]]), "agent_1": np.array([[1.0]])}
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None]

    def step(self, action):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        reward = {"agent_0": 1, "agent_1": 0}  # Constant reward
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        info = {}
        return observation, info


class ConstantRewardContActionsImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))}
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.0]]), "agent_1": np.array([[1.0]])}
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None]

    def step(self, action):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        reward = {"agent_0": 1, "agent_1": 0}  # Constant reward
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        info = {}
        return observation, info


class ObsDependentRewardEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[1]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2, 0.8]]), "agent_1": np.array([[0.8, 0.2]])},
            {"agent_0": np.array([[0.8, 0.2]]), "agent_1": np.array([[0.2, 0.8]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 1.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = self.last_obs
        reward = (
            {"agent_0": 1, "agent_1": 0}
            if self.last_obs["agent_0"] == 0
            else {"agent_0": 0, "agent_1": 1}
        )  # Reward depends on observation
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.array([0]), "agent_1": np.array([0])},
                {"agent_0": np.array([1]), "agent_1": np.array([1])},
            ]
        )
        info = {}
        return self.last_obs, info


class ObsDependentRewardImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2, 0.8]]), "agent_1": np.array([[0.8, 0.2]])},
            {"agent_0": np.array([[0.8, 0.2]]), "agent_1": np.array([[0.2, 0.8]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 1.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = self.last_obs
        reward = (
            {"agent_0": 1, "agent_1": 0}
            if np.mean(self.last_obs["agent_0"]) == 0
            else {"agent_0": 0, "agent_1": 1}
        )  # Reward depends on observation
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
            ]
        )
        info = {}
        return self.last_obs, info


class ObsDependentRewardContActionsEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[1]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2]]), "agent_1": np.array([[0.0]])},
            {"agent_0": np.array([[0.8]]), "agent_1": np.array([[0.6]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 1.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = self.last_obs
        reward = (
            {"agent_0": 1, "agent_1": 0}
            if self.last_obs["agent_0"] == 0
            else {"agent_0": 0, "agent_1": 1}
        )  # Reward depends on observation
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.array([0]), "agent_1": np.array([0])},
                {"agent_0": np.array([1]), "agent_1": np.array([1])},
            ]
        )
        info = {}
        return self.last_obs, info


class ObsDependentRewardContActionsImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2]]), "agent_1": np.array([[0.0]])},
            {"agent_0": np.array([[0.8]]), "agent_1": np.array([[0.6]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 1.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = self.last_obs
        reward = (
            {"agent_0": 1, "agent_1": 0}
            if np.mean(self.last_obs["agent_0"]) == 0
            else {"agent_0": 0, "agent_1": 1}
        )  # Reward depends on observation
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
            ]
        )
        info = {}
        return self.last_obs, info


class DiscountedRewardEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[1]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2, 0.8]]), "agent_1": np.array([[0.8, 0.2]])},
            {"agent_0": np.array([[0.8, 0.2]]), "agent_1": np.array([[0.2, 0.8]])},
        ]
        self.q_values = [
            {"agent_0": 0.99, "agent_1": 0.495},
            {"agent_0": 1.0, "agent_1": 0.5},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = {"agent_0": np.array([1]), "agent_1": np.array([1])}
        reward = (
            {"agent_0": 1, "agent_1": 0.5}
            if self.last_obs["agent_0"] == 1
            else {"agent_0": 0, "agent_1": 0}
        )  # Reward depends on observation  # Reward depends on observation
        terminated = self.last_obs  # Terminate after second step
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        self.last_obs = {"agent_0": np.array([1]), "agent_1": np.array([1])}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        info = {}
        return self.last_obs, info


class DiscountedRewardImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2, 0.8]]), "agent_1": np.array([[0.8, 0.2]])},
            {"agent_0": np.array([[0.8, 0.2]]), "agent_1": np.array([[0.2, 0.8]])},
        ]
        self.q_values = [
            {"agent_0": 0.99, "agent_1": 0.495},
            {"agent_0": 1.0, "agent_1": 0.5},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))}
        reward = (
            {"agent_0": 1, "agent_1": 0.5}
            if np.mean(self.last_obs["agent_0"]) == 1
            else {"agent_0": 0, "agent_1": 0}
        )  # Reward depends on observation  # Reward depends on observation
        terminated = {
            "agent_0": int(np.mean(self.last_obs["agent_0"])),
            "agent_1": int(np.mean(self.last_obs["agent_0"])),
        }  # Terminate after second step
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        self.last_obs = {
            "agent_0": np.ones((3, 32, 32)),
            "agent_1": np.ones((3, 32, 32)),
        }
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        info = {}
        return self.last_obs, info


class DiscountedRewardContActionsEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[1]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2]]), "agent_1": np.array([[0.4]])},
            {"agent_0": np.array([[0.8]]), "agent_1": np.array([[0.1]])},
        ]
        self.q_values = [
            {"agent_0": 0.99, "agent_1": 0.495},
            {"agent_0": 1.0, "agent_1": 0.5},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = {"agent_0": np.array([1]), "agent_1": np.array([1])}
        reward = (
            {"agent_0": 1, "agent_1": 0.5}
            if self.last_obs["agent_0"] == 1
            else {"agent_0": 0, "agent_1": 0}
        )  # Reward depends on observation  # Reward depends on observation
        terminated = self.last_obs  # Terminate after second step
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        self.last_obs = {"agent_0": np.array([1]), "agent_1": np.array([1])}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        info = {}
        return self.last_obs, info


class DiscountedRewardContActionsImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[0.2]]), "agent_1": np.array([[0.4]])},
            {"agent_0": np.array([[0.8]]), "agent_1": np.array([[0.1]])},
        ]
        self.q_values = [
            {"agent_0": 0.99, "agent_1": 0.495},
            {"agent_0": 1.0, "agent_1": 0.5},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [None, None]

    def step(self, action):
        observation = {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))}
        reward = (
            {"agent_0": 1, "agent_1": 0.5}
            if np.mean(self.last_obs["agent_0"]) == 1
            else {"agent_0": 0, "agent_1": 0}
        )  # Reward depends on observation  # Reward depends on observation
        terminated = {
            "agent_0": int(np.mean(self.last_obs["agent_0"])),
            "agent_1": int(np.mean(self.last_obs["agent_0"])),
        }  # Terminate after second step
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        self.last_obs = {
            "agent_0": np.ones((3, 32, 32)),
            "agent_1": np.ones((3, 32, 32)),
        }
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        info = {}
        return self.last_obs, info


class FixedObsPolicyEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(1),
            "agent_1": spaces.Discrete(1),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 1.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])}
        ]

    def step(self, action):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        reward = {
            "agent_0": [1, -1][action["agent_0"]],
            "agent_1": [-1, 1][action["agent_1"]],
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        info = {}
        return observation, info


class FixedObsPolicyImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 0.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 0.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 1.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])}
        ]

    def step(self, action):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        reward = {
            "agent_0": [1, -1][action["agent_0"]],
            "agent_1": [-1, 1][action["agent_1"]],
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        info = {}
        return observation, info


class FixedObsPolicyContActionsEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(1),
            "agent_1": spaces.Discrete(1),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0]]), "agent_1": np.array([[0.0]])},
        ]
        self.q_values = [
            {"agent_0": 0.0, "agent_1": 0.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [{"agent_0": np.array([1.0]), "agent_1": np.array([0.0])}]

    def step(self, action):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        reward = {
            "agent_0": -((1 - action["agent_0"]) ** 2),
            "agent_1": -((0 - action["agent_1"]) ** 2),
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        info = {}
        return observation, info


class FixedObsPolicyContActionsImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 0.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 0.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (1,)),
            "agent_1": spaces.Box(0.0, 1.0, (1,)),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0]]), "agent_1": np.array([[0.0]])},
        ]
        self.q_values = [
            {"agent_0": 0.0, "agent_1": 0.0}
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [{"agent_0": np.array([1.0]), "agent_1": np.array([0.0])}]

    def step(self, action):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        reward = {
            "agent_0": -((1 - action["agent_0"]) ** 2),
            "agent_1": -((0 - action["agent_1"]) ** 2),
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        observation = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        info = {}
        return observation, info


class PolicyEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 0.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 0.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]

    def step(self, action):
        observation = self.last_obs
        reward = {
            "agent_0": action["agent_0"] == self.last_obs["agent_0"],
            "agent_1": action["agent_1"] != self.last_obs["agent_1"],
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.array([0]), "agent_1": np.array([0])},
                {"agent_0": np.array([1]), "agent_1": np.array([1])},
                {"agent_0": np.array([0]), "agent_1": np.array([1])},
                {"agent_0": np.array([1]), "agent_1": np.array([0])},
            ]
        )
        info = {}
        return self.last_obs, info


class PolicyImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
        ]
        self.q_values = [
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 0.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 0.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]

    def step(self, action):
        observation = self.last_obs
        reward = {
            "agent_0": action["agent_0"] == np.mean(self.last_obs["agent_0"]),
            "agent_1": action["agent_1"] != np.mean(self.last_obs["agent_1"]),
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
            ]
        )
        info = {}
        return self.last_obs, info


class PolicyContActionsEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (2,)),
            "agent_1": spaces.Box(0.0, 1.0, (2,)),
        }
        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 0.0]]), "agent_1": np.array([[0.0, 0.0]])},
            {"agent_0": np.array([[1.0, 1.0]]), "agent_1": np.array([[1.0, 1.0]])},
        ]
        self.q_values = [
            {"agent_0": 0.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 0.0},
            {"agent_0": -2.0, "agent_1": -2.0},
            {"agent_0": -2.0, "agent_1": -2.0},
            {"agent_0": -1.0, "agent_1": -1.0},
            {"agent_0": -1.0, "agent_1": -1.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]

    def step(self, action):
        observation = self.last_obs
        reward = {}
        if self.last_obs["agent_0"]:  # last obs = 1, policy should be [0, 1]
            reward["agent_0"] = -((0 - action["agent_0"][0]) ** 2) - (
                (1 - action["agent_0"][1]) ** 2
            )
        else:  # last obs = 0, policy should be [1, 0]
            reward["agent_0"] = -((1 - action["agent_0"][0]) ** 2) - (
                (0 - action["agent_0"][1]) ** 2
            )
        if self.last_obs["agent_1"]:  # last obs = 1, policy should be [1, 0]
            reward["agent_1"] = -((1 - action["agent_1"][0]) ** 2) - (
                (0 - action["agent_1"][1]) ** 2
            )
        else:  # last obs = 0, policy should be [0, 1]
            reward["agent_1"] = -((0 - action["agent_1"][0]) ** 2) - (
                (1 - action["agent_1"][1]) ** 2
            )
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.array([0]), "agent_1": np.array([0])},
                {"agent_0": np.array([1]), "agent_1": np.array([1])},
                {"agent_0": np.array([0]), "agent_1": np.array([1])},
                {"agent_0": np.array([1]), "agent_1": np.array([0])},
            ]
        )
        info = {}
        return self.last_obs, info


class PolicyContActionsImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Box(0.0, 1.0, (2,)),
            "agent_1": spaces.Box(0.0, 1.0, (2,)),
        }
        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 0.0]]), "agent_1": np.array([[0.0, 0.0]])},
            {"agent_0": np.array([[1.0, 1.0]]), "agent_1": np.array([[1.0, 1.0]])},
        ]
        self.q_values = [
            {"agent_0": 0.0, "agent_1": 0.0},
            {"agent_0": 0.0, "agent_1": 0.0},
            {"agent_0": -2.0, "agent_1": -2.0},
            {"agent_0": -2.0, "agent_1": -2.0},
            {"agent_0": -1.0, "agent_1": -1.0},
            {"agent_0": -1.0, "agent_1": -1.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]

    def step(self, action):
        observation = self.last_obs
        reward = {}
        if np.mean(self.last_obs["agent_0"]):  # last obs = 1, policy should be [0, 1]
            reward["agent_0"] = -((0 - action["agent_0"][0]) ** 2) - (
                (1 - action["agent_0"][1]) ** 2
            )
        else:  # last obs = 0, policy should be [1, 0]
            reward["agent_0"] = -((1 - action["agent_0"][0]) ** 2) - (
                (0 - action["agent_0"][1]) ** 2
            )
        if np.mean(self.last_obs["agent_1"]):  # last obs = 1, policy should be [1, 0]
            reward["agent_1"] = -((1 - action["agent_1"][0]) ** 2) - (
                (0 - action["agent_1"][1]) ** 2
            )
        else:  # last obs = 0, policy should be [0, 1]
            reward["agent_1"] = -((0 - action["agent_1"][0]) ** 2) - (
                (1 - action["agent_1"][1]) ** 2
            )
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
            ]
        )
        info = {}
        return self.last_obs, info


class MultiPolicyEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {"agent_0": np.array([0]), "agent_1": np.array([0])}
        self.observation_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
            {"agent_0": np.array([[0]]), "agent_1": np.array([[1]])},
            {"agent_0": np.array([[1]]), "agent_1": np.array([[0]])},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
        ]
        self.q_values = [
            {"agent_0": 2.0, "agent_1": 2.0},
            {"agent_0": 2.0, "agent_1": 2.0},
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 0.0, "agent_1": 3.0},
            {"agent_0": 0.0, "agent_1": 3.0},
            {"agent_0": 3.0, "agent_1": 0.0},
            {"agent_0": 3.0, "agent_1": 0.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]

    def step(self, action):
        observation = self.last_obs
        reward = {
            "agent_0": 2 * (action["agent_0"] == self.last_obs["agent_0"])
            + (action["agent_1"] == self.last_obs["agent_1"]),
            "agent_1": 2 * (action["agent_1"] != self.last_obs["agent_1"])
            + (action["agent_0"] != self.last_obs["agent_0"]),
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.array([0]), "agent_1": np.array([0])},
                {"agent_0": np.array([1]), "agent_1": np.array([1])},
                {"agent_0": np.array([0]), "agent_1": np.array([1])},
                {"agent_0": np.array([1]), "agent_1": np.array([0])},
            ]
        )
        info = {}
        return self.last_obs, info


class MultiPolicyImageEnv:
    def __init__(self):
        self.possible_agents = ["agent_0", "agent_1"]
        self.agents = self.possible_agents
        self.max_num_agents = len(self.possible_agents)
        self.num_agents = len(self.agents)

        self.last_obs = {
            "agent_0": np.zeros((3, 32, 32)),
            "agent_1": np.zeros((3, 32, 32)),
        }
        self.observation_space = {
            "agent_0": spaces.Box(0.0, 1.0, (3, 32, 32)),
            "agent_1": spaces.Box(0.0, 1.0, (3, 32, 32)),
        }
        self.action_space = {
            "agent_0": spaces.Discrete(2),
            "agent_1": spaces.Discrete(2),
        }

        self.sample_obs = [
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
            {"agent_0": np.zeros((1, 3, 32, 32)), "agent_1": np.ones((1, 3, 32, 32))},
            {"agent_0": np.ones((1, 3, 32, 32)), "agent_1": np.zeros((1, 3, 32, 32))},
        ]
        self.sample_actions = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[1.0, 0.0]])},
        ]
        self.q_values = [
            {"agent_0": 2.0, "agent_1": 2.0},
            {"agent_0": 2.0, "agent_1": 2.0},
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 1.0, "agent_1": 1.0},
            {"agent_0": 0.0, "agent_1": 3.0},
            {"agent_0": 0.0, "agent_1": 3.0},
            {"agent_0": 3.0, "agent_1": 0.0},
            {"agent_0": 3.0, "agent_1": 0.0},
        ]  # Correct Q values to learn, s x a table
        self.policy_values = [
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
            {"agent_0": np.array([[1.0, 0.0]]), "agent_1": np.array([[1.0, 0.0]])},
            {"agent_0": np.array([[0.0, 1.0]]), "agent_1": np.array([[0.0, 1.0]])},
        ]

    def step(self, action):
        observation = self.last_obs
        reward = {
            "agent_0": 2
            * (np.mean(action["agent_0"]) == np.mean(self.last_obs["agent_0"]))
            + (np.mean(action["agent_1"]) == np.mean(self.last_obs["agent_1"])),
            "agent_1": 2
            * (np.mean(action["agent_1"]) != np.mean(self.last_obs["agent_1"]))
            + (np.mean(action["agent_0"]) != np.mean(self.last_obs["agent_0"])),
        }  # Reward depends on action
        terminated = {"agent_0": True, "agent_1": True}
        truncated = {"agent_0": False, "agent_1": False}
        info = {}
        return observation, reward, terminated, truncated, info

    def reset(self):
        self.last_obs = random.choice(
            [
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
                {"agent_0": np.zeros((3, 32, 32)), "agent_1": np.ones((3, 32, 32))},
                {"agent_0": np.ones((3, 32, 32)), "agent_1": np.zeros((3, 32, 32))},
            ]
        )
        info = {}
        return self.last_obs, info


def prepare_ma_states(states, one_hot, state_dims, device="cpu"):
    if one_hot:
        states = {
            agent_id: nn.functional.one_hot(
                torch.Tensor(state).long(), num_classes=state_dim[0]
            )
            .float()
            .squeeze(1)
            .to(device)
            for (agent_id, state), state_dim in zip(states.items(), state_dims)
        }
    else:
        states = {
            agent_id: torch.Tensor(state).to(device)
            for (agent_id, state) in states.items()
        }
    return states


def prepare_ma_actions(actions, device="cpu"):
    actions = {
        agent_id: torch.Tensor(action).to(device)
        for (agent_id, action) in actions.items()
    }
    return actions


def check_policy_q_learning_with_probe_env(
    env, algo_class, algo_args, memory, learn_steps=1000, device="cpu"
):
    print(f"Probe environment: {type(env).__name__}")

    agent = algo_class(**algo_args, device=device)

    state, _ = env.reset()
    for _ in range(10000):
        if agent.net_config["arch"] == "cnn":
            state = {agent_id: np.expand_dims(s, 0) for agent_id, s in state.items()}
        cont_actions, discrete_action = agent.getAction(state, epsilon=1)
        action = discrete_action if agent.discrete_actions else cont_actions
        next_state, reward, done, _, _ = env.step(action)
        if agent.net_config["arch"] == "cnn":
            state = {agent_id: np.squeeze(s) for agent_id, s in state.items()}
        memory.save2memory(state, cont_actions, reward, next_state, done)
        state = next_state
        if done[agent.agent_ids[0]]:
            state, _ = env.reset()

    # Learn from experiences
    for _ in trange(learn_steps):
        experiences = memory.sample(agent.batch_size)
        # Learn according to agent's RL algorithm
        agent.learn(experiences)

    with torch.no_grad():
        for agent_id, actor, critic in zip(
            agent.agent_ids, agent.actors, agent.critics
        ):
            for sample_obs, sample_action, q_values, policy_values in zip(
                env.sample_obs, env.sample_actions, env.q_values, env.policy_values
            ):
                state = prepare_ma_states(
                    sample_obs, agent.one_hot, agent.state_dims, device
                )

                if q_values is not None:
                    action = prepare_ma_actions(sample_action, device)
                    if agent.arch == "mlp":
                        input_combined = torch.cat(
                            list(state.values()) + list(action.values()), 1
                        )
                        predicted_q_values = (
                            critic(input_combined).detach().cpu().numpy()[0]
                        )
                    else:
                        stacked_states = torch.stack(list(state.values()), dim=2)
                        stacked_actions = torch.cat(list(action.values()), dim=1)
                        predicted_q_values = (
                            critic(stacked_states, stacked_actions)
                            .detach()
                            .cpu()
                            .numpy()[0]
                        )
                    # print("---")
                    # print(agent_id, "q", q_values[agent_id], predicted_q_values)
                    # assert np.allclose(q_values[agent_id], predicted_q_values, atol=0.1):
                    if not np.allclose(
                        q_values[agent_id], predicted_q_values, atol=0.1
                    ):
                        print(agent_id, "q", q_values[agent_id], predicted_q_values)

                if policy_values is not None:
                    if agent.arch == "mlp":
                        predicted_policy_values = (
                            actor(state[agent_id]).detach().cpu().numpy()[0]
                        )
                    else:
                        predicted_policy_values = (
                            actor(state[agent_id].unsqueeze(2))
                            .detach()
                            .cpu()
                            .numpy()[0]
                        )
                    # print(agent_id, "pol", policy_values[agent_id], predicted_policy_values)
                    # assert np.allclose(policy_values[agent_id], predicted_policy_values, atol=0.1)
                    if not np.allclose(
                        policy_values[agent_id], predicted_policy_values, atol=0.1
                    ):
                        print(
                            agent_id,
                            "pol",
                            policy_values[agent_id],
                            predicted_policy_values,
                        )


# if __name__ == "__main__":
#     from agilerl.algorithms.maddpg import MADDPG
#     from agilerl.components.multi_agent_replay_buffer import MultiAgentReplayBuffer

#     device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

#     vector_envs = [
#         (ConstantRewardEnv(), 1000),
#         (ObsDependentRewardEnv(), 1000),
#         (DiscountedRewardEnv(), 3000),
#         (FixedObsPolicyEnv(), 1000),
#         (PolicyEnv(), 4000),
#         (MultiPolicyEnv(), 8000),
#     ]

#     for env, learn_steps in vector_envs:
#         algo_args = {
#             "state_dims": [(env.observation_space[agent].n,) for agent in env.agents],
#             "action_dims": [env.action_space[agent].n for agent in env.agents],
#             "one_hot": True,
#             "n_agents": env.num_agents,
#             "agent_ids": env.possible_agents,
#             "max_action": [(1.0,), (1.0,)],
#             "min_action": [(0.0,), (0.0,)],
#             "discrete_actions": True,
#             "net_config": {"arch": "mlp", "hidden_size": [32, 32]},
#             "batch_size": 256,
#         }
#         field_names = ["state", "action", "reward", "next_state", "done"]
#         memory = MultiAgentReplayBuffer(
#             memory_size=10000,  # Max replay buffer size
#             field_names=field_names,  # Field names to store in memory
#             agent_ids=algo_args["agent_ids"],
#             device=device,
#         )

#         check_policy_q_learning_with_probe_env(env, MADDPG, algo_args, memory, learn_steps, device)

#     image_envs = [
#         (ConstantRewardImageEnv(), 2000),
#         (ObsDependentRewardImageEnv(), 2000),
#         (DiscountedRewardImageEnv(), 5000),
#         (FixedObsPolicyImageEnv(), 5000),
#         (PolicyImageEnv(), 10000),
#         (MultiPolicyImageEnv(), 15000),
#     ]

#     for env, learn_steps in image_envs:
#         algo_args = {
#             "state_dims": [(env.observation_space[agent].shape) for agent in env.agents],
#             "action_dims": [env.action_space[agent].n for agent in env.agents],
#             "one_hot": False,
#             "n_agents": env.num_agents,
#             "agent_ids": env.possible_agents,
#             "max_action": [(1.0,), (1.0,)],
#             "min_action": [(0.0,), (0.0,)],
#             "discrete_actions": True,
#             "net_config": {
#                 "arch": "cnn",  # Network architecture
#                 "hidden_size": [32],  # Network hidden size
#                 "channel_size": [32, 32],  # CNN channel size
#                 "kernel_size": [(1, 4, 4), (1, 3, 3)],  # CNN kernel size
#                 "stride_size": [4, 2],  # CNN stride size
#                 "normalize": False,  # Normalize image from range [0,255] to [0,1]
#             },
#             "lr_actor": 1e-5,
#             "lr_critic": 1e-3,
#             "batch_size": 128,
#         }
#         field_names = ["state", "action", "reward", "next_state", "done"]
#         memory = MultiAgentReplayBuffer(
#             memory_size=10000,  # Max replay buffer size
#             field_names=field_names,  # Field names to store in memory
#             agent_ids=algo_args["agent_ids"],
#             device=device,
#         )

#         check_policy_q_learning_with_probe_env(env, MADDPG, algo_args, memory, learn_steps, device)

#     cont_vector_envs = [
#         (ConstantRewardContActionsEnv(), 2000),
#         (ObsDependentRewardContActionsEnv(), 2000),
#         (DiscountedRewardContActionsEnv(), 5000),
#         (FixedObsPolicyContActionsEnv(), 3000),
#         (PolicyContActionsEnv(), 5000),
#     ]

#     for env, learn_steps in cont_vector_envs:
#         algo_args = {
#             "state_dims": [(env.observation_space[agent].n,) for agent in env.agents],
#             "action_dims": [env.action_space[agent].shape[0] for agent in env.agents],
#             "one_hot": True,
#             "n_agents": env.num_agents,
#             "agent_ids": env.possible_agents,
#             "max_action": [(1.0,), (1.0,)],
#             "min_action": [(0.0,), (0.0,)],
#             "discrete_actions": False,
#             "net_config": {"arch": "mlp", "hidden_size": [32, 32]},
#             "batch_size": 256,
#         }
#         field_names = ["state", "action", "reward", "next_state", "done"]
#         memory = MultiAgentReplayBuffer(
#             memory_size=10000,  # Max replay buffer size
#             field_names=field_names,  # Field names to store in memory
#             agent_ids=algo_args["agent_ids"],
#             device=device,
#         )

#         check_policy_q_learning_with_probe_env(env, MADDPG, algo_args, memory, learn_steps, device)

#     cont_image_envs = [
#         (ConstantRewardContActionsImageEnv(), 2000),
#         (ObsDependentRewardContActionsImageEnv(), 2000),
#         (DiscountedRewardContActionsImageEnv(), 3000),
#         (FixedObsPolicyContActionsImageEnv(), 5000),
#         (PolicyContActionsImageEnv(), 10000),
#     ]

#     for env, learn_steps in cont_image_envs:
#         algo_args = {
#             "state_dims": [(env.observation_space[agent].shape) for agent in env.agents],
#             "action_dims": [env.action_space[agent].shape[0] for agent in env.agents],
#             "one_hot": False,
#             "n_agents": env.num_agents,
#             "agent_ids": env.possible_agents,
#             "max_action": [(1.0,), (1.0,)],
#             "min_action": [(0.0,), (0.0,)],
#             "discrete_actions": False,
#             "net_config": {
#                 "arch": "cnn",  # Network architecture
#                 "hidden_size": [32],  # Network hidden size
#                 "channel_size": [32, 32],  # CNN channel size
#                 "kernel_size": [(1, 4, 4), (1, 3, 3)],  # CNN kernel size
#                 "stride_size": [4, 2],  # CNN stride size
#                 "normalize": False,  # Normalize image from range [0,255] to [0,1]
#             },
#             "lr_actor": 1e-4,
#             "lr_critic": 1e-3,
#             "batch_size": 128,
#         }
#         field_names = ["state", "action", "reward", "next_state", "done"]
#         memory = MultiAgentReplayBuffer(
#             memory_size=10000,  # Max replay buffer size
#             field_names=field_names,  # Field names to store in memory
#             agent_ids=algo_args["agent_ids"],
#             device=device,
#         )

#         check_policy_q_learning_with_probe_env(env, MADDPG, algo_args, memory, learn_steps, device)



"""
Author: Burak M Gonultas
https://github.com/gonultasbu
---
Original Reference:
https://github.com/Farama-Foundation/SuperSuit/issues/43#issuecomment-751792111
"""

from multiprocessing import Pipe, Process

import numpy as np


def worker(remote, parent_remote, env_fn_wrapper):
    """Worker class

    References:
        https://github.com/openai/baselines/tree/master/baselines/common/vec_env
    """
    parent_remote.close()
    env = env_fn_wrapper.x()
    while True:
        cmd, data = remote.recv()
        if cmd == "step":
            data = {
                possible_agent: np.array(data[idx]).squeeze()
                for idx, possible_agent in enumerate(env.possible_agents)
            }
            ob, reward, dones, truncs, info = env.step(data)
            ob = list(ob.values())
            reward = list(reward.values())
            dones = list(dones.values())
            info = list(info.values())
            truncs = list(truncs.values())
            remote.send((ob, reward, dones, truncs, info))
        elif cmd == "reset":
            ob, infos = env.reset(seed=data, options=None)
            ob = list(ob.values())
            infos = list(infos.values())
            remote.send((ob, infos))
        elif cmd == "close":
            env.close()
            remote.close()
            break
        elif cmd == "seed":
            env.seed(data)
        elif cmd == "render":
            env.render()
        else:
            raise NotImplementedError


class VecEnv:
    """An abstract asynchronous, vectorized environment

    References:
        https://github.com/openai/baselines/tree/master/baselines/common/vec_env
    """

    def __init__(self, num_envs, possible_agents):
        self.num_envs = num_envs
        self.agents = possible_agents
        self.num_agents = len(self.agents)

    def reset(self, seed=None, options=None):
        """
        Reset all the environments and return an array of
        observations, or a tuple of observation arrays.
        If step_async is still doing work, that work will
        be cancelled and step_wait() should not be called
        until step_async() is invoked again.
        """
        pass

    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is
        already pending.
        """
        pass

    def step_wait(self):
        """
        Wait for the step taken with step_async().
        Returns (obs, rews, dones, infos):
         - obs: an array of observations, or a tuple of
                arrays of observations.
         - rews: an array of rewards
         - dones: an array of "episode done" booleans
         - infos: a sequence of info objects
        """
        pass

    def close(self):
        """
        Clean up the environments' resources.
        """
        pass

    def step(self, actions):
        passed_actions_list = [[] for _ in list(actions.values())[0]]
        for env_idx, _ in enumerate(list(actions.values())[0]):
            for possible_agent in self.agents:
                passed_actions_list[env_idx].append(actions[possible_agent][env_idx])
        self.step_async(passed_actions_list)
        return self.step_wait()


class CloudpickleWrapper:
    """Uses cloudpickle to serialize contents
    (otherwise multiprocessing tries to use pickle)

    References:
        https://github.com/openai/baselines/tree/master/baselines/common/vec_env
    """

    def __init__(self, x):
        self.x = x

    def __getstate__(self):
        import cloudpickle

        return cloudpickle.dumps(self.x)

    def __setstate__(self, ob):
        import pickle

        self.x = pickle.loads(ob)


class SubprocVecEnv(VecEnv):
    """Vectorized environment class that collects samples in parallel using subprocesses

    Args:
        env_fns (list): list of gym environments to run in subprocesses

    References:
        https://github.com/openai/baselines/tree/master/baselines/common/vec_env
    """

    def __init__(self, env_fns):
        self.env = env_fns[0]()
        self.waiting = False
        self.observation_space = self.env.observation_space
        self.action_space = self.env.action_space
        self.closed = False
        self.nenvs = len(env_fns)
        self.remotes, self.work_remotes = zip(*[Pipe() for _ in range(self.nenvs)])
        self.ps = [
            Process(
                target=worker, args=(work_remote, remote, CloudpickleWrapper(env_fn))
            )
            for (work_remote, remote, env_fn) in zip(
                self.work_remotes, self.remotes, env_fns
            )
        ]
        for p in self.ps:
            p.daemon = (
                True  # If the main process crashes, we should not cause things to hang
            )
            p.start()
        for remote in self.work_remotes:
            remote.close()
        VecEnv.__init__(
            self,
            len(env_fns),
            self.env.possible_agents,
        )

    def seed(self, value):
        for i_remote, remote in enumerate(self.remotes):
            remote.send(("seed", value + i_remote))

    def step_async(self, actions):
        for remote, action in zip(self.remotes, actions):
            remote.send(("step", action))
        self.waiting = True

    def step_wait(self):
        results = [remote.recv() for remote in self.remotes]
        self.waiting = False
        obs, rews, dones, truncs, infos = zip(*results)

        ret_obs_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        ret_rews_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        ret_dones_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        ret_truncs_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        ret_infos_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        for env_idx, _ in enumerate(obs):
            for agent_idx, possible_agent in enumerate(self.env.possible_agents):
                ret_obs_dict[possible_agent].append(obs[env_idx][agent_idx])
                ret_rews_dict[possible_agent].append(rews[env_idx][agent_idx])
                ret_dones_dict[possible_agent].append(dones[env_idx][agent_idx])
                ret_truncs_dict[possible_agent].append(truncs[env_idx][agent_idx])
                ret_infos_dict[possible_agent].append(infos[env_idx][agent_idx])

        for agent_idx, possible_agent in enumerate(self.env.possible_agents):
            for op_dict in [
                ret_obs_dict,
                ret_rews_dict,
                ret_dones_dict,
                ret_truncs_dict,
                ret_infos_dict,
            ]:
                op_dict[possible_agent] = np.stack(op_dict[possible_agent])
        return (
            ret_obs_dict,
            ret_rews_dict,
            ret_dones_dict,
            ret_truncs_dict,
            ret_infos_dict,
        )

    def reset(self, seed=None, options=None):
        for remote in self.remotes:
            remote.send(("reset", seed))
        results = [remote.recv() for remote in self.remotes]
        obs, infos = zip(*results)
        ret_obs_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        ret_infos_dict = {
            possible_agent: []
            for idx, possible_agent in enumerate(self.env.possible_agents)
        }
        for env_idx, _ in enumerate(obs):
            for agent_idx, possible_agent in enumerate(self.env.possible_agents):
                ret_obs_dict[possible_agent].append(obs[env_idx][agent_idx])
                ret_infos_dict[possible_agent].append(infos[env_idx][agent_idx])
        for agent_idx, possible_agent in enumerate(self.env.possible_agents):
            for op_dict in [
                ret_obs_dict,
                ret_infos_dict,
            ]:
                op_dict[possible_agent] = np.stack(op_dict[possible_agent])
        return (ret_obs_dict, ret_infos_dict)

    def render(self):
        self.remotes[0].send(("render", None))

    def close(self):
        if self.closed:
            return
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
        for remote in self.remotes:
            remote.send(("close", None))
        for p in self.ps:
            p.join()
            self.closed = True

    def sample_personas(self, is_train, is_val=True, path="./"):
        return self.env.sample_personas(is_train=is_train, is_val=is_val, path=path)



import os
import warnings
from datetime import datetime

import numpy as np
import wandb
from tqdm import trange

from agilerl.utils.utils import calculate_vectorized_scores

os.environ["CUDA_LAUNCH_BLOCKING"] = "1"


def train_on_policy(
    env,
    env_name,
    algo,
    pop,
    INIT_HP=None,
    MUT_P=None,
    swap_channels=False,
    n_episodes=2000,
    max_steps=500,
    evo_epochs=5,
    evo_loop=1,
    target=200.0,
    tournament=None,
    mutation=None,
    checkpoint=None,
    checkpoint_path=None,
    save_elite=False,
    elite_path=None,
    wb=False,
    verbose=True,
    accelerator=None,
    wandb_api_key=None,
):
    """The general on-policy RL training function. Returns trained population of agents
    and their fitnesses.

    :param env: The environment to train in. Can be vectorized.
    :type env: Gym-style environment
    :param env_name: Environment name
    :type env_name: str
    :param algo: RL algorithm name
    :type algo: str
    :param pop: Population of agents
    :type pop: list[object]
    :param INIT_HP: Dictionary containing initial hyperparameters, defaults to None
    :type INIT_HP: dict, optional
    :param MUT_P: Dictionary containing mutation parameters, defaults to None
    :type MUT_P: dict, optional
    :param swap_channels: Swap image channels dimension from last to first
        [H, W, C] -> [C, H, W], defaults to False
    :type swap_channels: bool, optional
    :param n_episodes: Maximum number of training episodes, defaults to 2000
    :type n_episodes: int, optional
    :param max_steps: Maximum number of steps in environment per episode, defaults to
        500
    :type max_steps: int, optional
    :param evo_epochs: Evolution frequency (episodes), defaults to 5
    :type evo_epochs: int, optional
    :param evo_loop: Number of evaluation episodes, defaults to 1
    :type evo_loop: int, optional
    :param target: Target score for early stopping, defaults to 200.
    :type target: float, optional
    :param tournament: Tournament selection object, defaults to None
    :type tournament: object, optional
    :param mutation: Mutation object, defaults to None
    :type mutation: object, optional
    :param checkpoint: Checkpoint frequency (episodes), defaults to None
    :type checkpoint: int, optional
    :param checkpoint_path: Location to save checkpoint, defaults to None
    :type checkpoint_path: str, optional
    :param save_elite: Boolean flag indicating whether to save elite member at the end
        of training, defaults to False
    :type save_elite: bool, optional
    :param elite_path: Location to save elite agent, defaults to None
    :type elite_path: str, optional
    :param wb: Weights & Biases tracking, defaults to False
    :type wb: bool, optional
    :param verbose: Display training stats, defaults to True
    :type verbose: bool, optional
    :param accelerator: Accelerator for distributed computing, defaults to None
    :type accelerator: accelerate.Accelerator(), optional
    :param wandb_api_key: API key for Weights & Biases, defaults to None
    :type wandb_api_key: str, optional
    """
    assert isinstance(
        algo, str
    ), "'algo' must be the name of the algorithm as a string."
    assert isinstance(n_episodes, int), "Number of episodes must be an integer."
    assert isinstance(max_steps, int), "Number of steps must be an integer."
    assert isinstance(evo_epochs, int), "Evolution frequency must be an integer."
    if target is not None:
        assert isinstance(
            target, (float, int)
        ), "Target score must be a float or an integer."
    if checkpoint is not None:
        assert isinstance(checkpoint, int), "Checkpoint must be an integer."
    assert isinstance(
        wb, bool
    ), "'wb' must be a boolean flag, indicating whether to record run with W&B"
    assert isinstance(verbose, bool), "Verbose must be a boolean."
    if save_elite is False and elite_path is not None:
        warnings.warn(
            "'save_elite' set to False but 'elite_path' has been defined, elite will not\
                      be saved unless 'save_elite' is set to True."
        )
    if checkpoint is None and checkpoint_path is not None:
        warnings.warn(
            "'checkpoint' set to None but 'checkpoint_path' has been defined, checkpoint will not\
                      be saved unless 'checkpoint' is defined."
        )

    if wb:
        if not hasattr(wandb, "api"):
            if wandb_api_key is not None:
                wandb.login(key=wandb_api_key)
            else:
                warnings.warn("Must login to wandb with API key.")

        config_dict = {}
        if INIT_HP is not None:
            config_dict.update(INIT_HP)
        if MUT_P is not None:
            config_dict.update(MUT_P)

        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.is_main_process:
                wandb.init(
                    # set the wandb project where this run will be logged
                    project="AgileRL",
                    name="{}-EvoHPO-{}-{}".format(
                        env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
                    ),
                    # track hyperparameters and run metadata
                    config=config_dict,
                )
            accelerator.wait_for_everyone()
        else:
            wandb.init(
                # set the wandb project where this run will be logged
                project="AgileRL",
                name="{}-EvoHPO-{}-{}".format(
                    env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
                ),
                # track hyperparameters and run metadata
                config=config_dict,
            )

    if accelerator is not None:
        accel_temp_models_path = f"models/{env_name}"
        if accelerator.is_main_process:
            if not os.path.exists(accel_temp_models_path):
                os.makedirs(accel_temp_models_path)

    # Detect if environment is vectorised
    if hasattr(env, "num_envs"):
        is_vectorised = True
    else:
        is_vectorised = False

    save_path = (
        checkpoint_path.split(".pt")[0]
        if checkpoint_path is not None
        else "{}-EvoHPO-{}-{}".format(
            env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
        )
    )

    if accelerator is not None:
        print(f"\nDistributed training on {accelerator.device}...")
    else:
        print("\nTraining...")

    bar_format = "{l_bar}{bar:10}| {n:4}/{total_fmt} [{elapsed:>7}<{remaining:>7}, {rate_fmt}{postfix}]"
    if accelerator is not None:
        pbar = trange(
            n_episodes,
            unit="ep",
            bar_format=bar_format,
            ascii=True,
            dynamic_ncols=True,
            disable=not accelerator.is_local_main_process,
        )
    else:
        pbar = trange(
            n_episodes, unit="ep", bar_format=bar_format, ascii=True, dynamic_ncols=True
        )

    pop_loss = [[] for _ in pop]
    pop_fitnesses = []
    total_steps = 0
    loss = None

    # Pre-training mutation
    if accelerator is not None:
        if mutation is not None:
            pop = mutation.mutation(pop, pre_training_mut=True)

    # RL training loop
    for idx_epi in pbar:
        if accelerator is not None:
            accelerator.wait_for_everyone()
        for agent_idx, agent in enumerate(pop):  # Loop through population
            state = env.reset()[0]  # Reset environment at start of episode
            score = 0

            states = []
            actions = []
            log_probs = []
            rewards = []
            dones = []
            values = []
            truncs = []

            for idx_step in range(max_steps):
                if swap_channels:
                    state = np.moveaxis(state, [-1], [-3])
                # Get next action from agent
                action, log_prob, _, value = agent.getAction(state)

                if not is_vectorised:
                    action = action[0]
                    log_prob = log_prob[0]
                    value = value[0]
                next_state, reward, done, trunc, _ = env.step(
                    action
                )  # Act in environment

                states.append(state)
                actions.append(action)
                log_probs.append(log_prob)
                rewards.append(reward)
                dones.append(done)
                values.append(value)
                truncs.append(trunc)

                state = next_state
                score += reward

            if swap_channels:
                next_state = np.moveaxis(next_state, [-1], [-3])

            if is_vectorised:
                scores = calculate_vectorized_scores(
                    np.array(rewards).transpose((1, 0)),
                    np.array(dones).transpose((1, 0)),
                )
                score = np.mean(scores)

            agent.scores.append(score)

            experiences = (
                states,
                actions,
                log_probs,
                rewards,
                dones,
                values,
                next_state,
            )
            # Learn according to agent's RL algorithm
            loss = agent.learn(experiences)
            pop_loss[agent_idx].append(loss)
            agent.steps[-1] += idx_step + 1
            total_steps += idx_step + 1

        # Now evolve if necessary
        if (idx_epi + 1) % evo_epochs == 0:
            # Evaluate population
            fitnesses = [
                agent.test(
                    env, swap_channels=swap_channels, max_steps=max_steps, loop=evo_loop
                )
                for agent in pop
            ]
            pop_fitnesses.append(fitnesses)
            mean_scores = np.mean([agent.scores[-evo_epochs:] for agent in pop], axis=1)

            wandb_dict = {
                "global_step": (
                    total_steps * accelerator.state.num_processes
                    if accelerator is not None and accelerator.is_main_process
                    else total_steps
                ),
                "train/mean_score": np.mean(mean_scores),
                "eval/mean_fitness": np.mean(fitnesses),
                "eval/best_fitness": np.max(fitnesses),
            }

            agent_loss_dict = {
                f"train/agent_{index}_loss": np.mean(loss_[-evo_epochs:])
                for index, loss_ in enumerate(pop_loss)
            }
            wandb_dict.update(agent_loss_dict)

            if wb:
                if accelerator is not None:
                    accelerator.wait_for_everyone()
                    if accelerator.is_main_process:
                        wandb.log(wandb_dict)
                    accelerator.wait_for_everyone()
                else:
                    wandb.log(wandb_dict)

            # Update step counter
            for agent in pop:
                agent.steps.append(agent.steps[-1])

            # Early stop if consistently reaches target
            if (
                np.all(
                    np.greater([np.mean(agent.fitness[-100:]) for agent in pop], target)
                )
                and idx_epi >= 100
            ):
                if wb:
                    wandb.finish()
                return pop, pop_fitnesses

            # Tournament selection and population mutation
            if tournament and mutation is not None:
                if accelerator is not None:
                    accelerator.wait_for_everyone()
                    for model in pop:
                        model.unwrap_models()
                    accelerator.wait_for_everyone()
                    if accelerator.is_main_process:
                        elite, pop = tournament.select(pop)
                        pop = mutation.mutation(pop)
                        for pop_i, model in enumerate(pop):
                            model.saveCheckpoint(
                                f"{accel_temp_models_path}/{algo}_{pop_i}.pt"
                            )
                    accelerator.wait_for_everyone()
                    if not accelerator.is_main_process:
                        for pop_i, model in enumerate(pop):
                            model.loadCheckpoint(
                                f"{accel_temp_models_path}/{algo}_{pop_i}.pt"
                            )
                    accelerator.wait_for_everyone()
                    for model in pop:
                        model.wrap_models()
                else:
                    elite, pop = tournament.select(pop)
                    pop = mutation.mutation(pop)

                if save_elite and (idx_epi + 1 == n_episodes):
                    elite_save_path = (
                        elite_path.split(".pt")[0]
                        if elite_path is not None
                        else "{}-elite_{}-{}".format(
                            env_name, algo, datetime.now().strftime("%m%d%Y%H%M%S")
                        )
                    )
                    elite.saveCheckpoint(f"{elite_save_path}.pt")

            if verbose:
                fitness = ["%.2f" % fitness for fitness in fitnesses]
                avg_fitness = ["%.2f" % np.mean(agent.fitness[-100:]) for agent in pop]
                avg_score = ["%.2f" % np.mean(agent.scores[-100:]) for agent in pop]
                agents = [agent.index for agent in pop]
                num_steps = [agent.steps[-1] for agent in pop]
                muts = [agent.mut for agent in pop]
                pbar.update(0)

                print(
                    f"""
                    --- Epoch {idx_epi + 1} ---
                    Fitness:\t\t{fitness}
                    100 fitness avgs:\t{avg_fitness}
                    100 score avgs:\t{avg_score}
                    Agents:\t\t{agents}
                    Steps:\t\t{num_steps}
                    Mutations:\t\t{muts}
                    """,
                    end="\r",
                )

        # Save model checkpoint
        if checkpoint is not None:
            print("Checkpoint")
            if (idx_epi + 1) % checkpoint == 0:
                if accelerator is not None:
                    accelerator.wait_for_everyone()
                    for model in pop:
                        model.unwrap_models()
                    accelerator.wait_for_everyone()
                    if accelerator.is_main_process:
                        for i, agent in enumerate(pop):
                            agent.saveCheckpoint(f"{save_path}_{i}_{idx_epi+1}.pt")
                        print("Saved checkpoint.")
                    accelerator.wait_for_everyone()
                    for model in pop:
                        model.wrap_models()
                    accelerator.wait_for_everyone()
                else:
                    for i, agent in enumerate(pop):
                        agent.saveCheckpoint(f"{save_path}_{i}_{idx_epi+1}.pt")
                    print("Saved checkpoint.")

    if wb:
        if accelerator is not None:
            accelerator.wait_for_everyone()
            if accelerator.is_main_process:
                wandb.finish()
            accelerator.wait_for_everyone()
        else:
            wandb.finish()

    return pop, pop_fitnesses
